
#include "device_launch_parameters.h"
#include "cuda_runtime.h"
#include "helper_cuda.h"

#include <cmath>

#include "Matrix.h"

// Thread block size
#define BLOCK_SIZE 16

// Forward declaration of the matrix multiplication kernel
__global__ void MatMulKernel(const Matrix, const Matrix, Matrix);


// Matrix multiplication - Host code
// Matrix dimensions are assumed to be multiples of BLOCK_SIZE
void MatMul(const Matrix A, const Matrix B, Matrix C)
{
    // Load A and B to device memory
    Matrix d_A;
    d_A.width = A.width; d_A.height = A.height;

    checkCudaErrors( cudaMallocPitch( &d_A.elements,&d_A.stride_in_bytes, d_A.width*sizeof(float),d_A.height ) );
    checkCudaErrors( cudaMemcpy2D( d_A.elements,d_A.stride_in_bytes,
                                   A.elements,A.stride_in_bytes,A.width*sizeof(float), A.height,
                                   cudaMemcpyHostToDevice ) );
    Matrix d_B;
    d_B.width = B.width; d_B.height = B.height;

    checkCudaErrors( cudaMallocPitch( &d_B.elements,&d_B.stride_in_bytes, d_B.width*sizeof(float),d_B.height ) );
    checkCudaErrors( cudaMemcpy2D( d_B.elements,d_B.stride_in_bytes,
                                   B.elements,B.stride_in_bytes,B.width*sizeof(float), B.height,
                                 cudaMemcpyHostToDevice ) );

    // Allocate C in device memory
    Matrix d_C;
    d_C.width = C.width; d_C.height = C.height;
    
    checkCudaErrors( cudaMallocPitch( &d_C.elements,&d_C.stride_in_bytes, d_C.width*sizeof(float),d_C.height ) );
    checkCudaErrors( cudaMemcpy2D( d_C.elements,d_C.stride_in_bytes,
                                   C.elements,C.stride_in_bytes,C.width*sizeof(float), C.height,
                                   cudaMemcpyHostToDevice ) );

    cudaEvent_t start, stop;

    checkCudaErrors( cudaEventCreate( &start ) );
    checkCudaErrors( cudaEventCreate( &stop ) );

    // Invoke kernel
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
    dim3 dimGrid(B.width / dimBlock.x, A.height / dimBlock.y);

    checkCudaErrors( cudaEventRecord( start ) );

    MatMulKernel<<<dimGrid, dimBlock>>>(d_A, d_B, d_C);

    checkCudaErrors( cudaGetLastError() );
    checkCudaErrors( cudaEventRecord( stop ) );

    // Wait for the stop event
    checkCudaErrors( cudaEventSynchronize( stop ) );
    float elapsedTime = 0.0f;
    checkCudaErrors( cudaEventElapsedTime( &elapsedTime, start, stop ) );
    printf( "Time to compute:  %3.1f ms\n", elapsedTime );

    checkCudaErrors( cudaEventDestroy( start ) );
    checkCudaErrors( cudaEventDestroy( stop ) );

    // Read C from device memory
    checkCudaErrors( cudaMemcpy2D( C.elements,C.stride_in_bytes, d_C.elements,d_C.stride_in_bytes, d_C.width * sizeof(float), d_C.height,
                                 cudaMemcpyDeviceToHost ) );

    // Free device memory
    checkCudaErrors( cudaFree( d_A.elements ) );
    checkCudaErrors( cudaFree( d_B.elements ) );
    checkCudaErrors( cudaFree( d_C.elements ) );
}

// Get a matrix element
__device__ float GetElement(const Matrix A, int row, int col)
{
    return *((float*)((char*)A.elements + row * A.stride_in_bytes) + col);
}

// Set a matrix element
__device__ void SetElement(Matrix A, int row, int col,
                            float value)
{
    *((float*)((char*)A.elements + row * A.stride_in_bytes) + col) = value;
}


// Matrix multiplication kernel called by MatMul()
__global__ void MatMulKernel(Matrix A, Matrix B, Matrix C)
{
    // Each thread computes one element of C
    // by accumulating results into Cvalue
    float Cvalue = 0;
    int row = blockIdx.y * blockDim.y + threadIdx.y;
    int col = blockIdx.x * blockDim.x + threadIdx.x;
    for ( int e = 0; e < A.width; ++e )
        Cvalue += GetElement( A, row, e )
        * GetElement(B,e, col);
    SetElement( C, row, col, Cvalue );
}



int main()
{
    Matrix h_A, h_B, h_C_gold;
    int a_rows = 64 * BLOCK_SIZE, a_cols = 32 * BLOCK_SIZE, b_cols = 16 * BLOCK_SIZE;

    CreateMatrix( h_A, a_rows, a_cols );
    CreateMatrix( h_B, a_cols, b_cols );
    CreateMatrix( h_C_gold, a_rows, b_cols );

    FillRandomMatrix( h_A );
    FillRandomMatrix( h_B );

    MatMul_CPU( h_A, h_B, h_C_gold );

    Matrix h_C;
    CreateMatrix( h_C, a_rows, b_cols );
    MatMul( h_A, h_B, h_C );

    float L2norm = 0;
    for ( int i = 0; i < h_C.height; i++ )
    {
        for ( int j = 0; j < h_C.width; j++ )
        {
            float diff = h_C.elements[ i * h_C.width + j ] - h_C_gold.elements[ i * h_C_gold.width + j ];
            L2norm += diff * diff;
        }
    }

    L2norm = sqrtf( L2norm / float(h_C.width * h_C.height) );
    if (L2norm > 1e-4f )
    {
        printf( "Mismatch\n" );
    }

    DestroyMatrix( h_A );
    DestroyMatrix( h_B );
    DestroyMatrix( h_C_gold );
    DestroyMatrix( h_C );

    return 0;
}
